#############################################################################################################################
#### No Polaris, no problem - polar alignment
#### Norman Cleesattel
#### Version 06. Jan 2021

import clr
import csv
import math
import os
import shlex
import subprocess
import sys
import time

clr.AddReference("SharpCap")
clr.AddReference('System.Windows.Forms')
clr.AddReference('System.Drawing')
clr.AddReference("ASCOM.DriverAccess")

from SharpCap.UI import CaptureLimitType
from System.Windows.Forms import *            
from System.Drawing import Pen, Brush, Point, Color, Font, FontStyle, Size, Image, Icon, Rectangle, Brushes, SolidBrush, Pens
from System.Drawing.Drawing2D import GraphicsPath, CustomLineCap, SmoothingMode, LineJoin
from System.Environment import GetFolderPath, SpecialFolder
from System.Threading import Thread
from ASCOM.DriverAccess import *


desktop_folder = GetFolderPath(SpecialFolder.Desktop)
user_doc_folder = GetFolderPath(SpecialFolder.MyDocuments)

default_pracma_part1 = user_doc_folder
default_pracma_part1 = default_pracma_part1.replace('\\', '\\\\')

#############################################################################################################################
#############################################################################################################################
#### Absolute PATHS
#### change if your system requires it

default_astap = r"C:\\Progra~1\\astap\\astap.exe"
default_Rscript = r"C:\\Progra~1\\R\\R-4.0.3\\bin\\Rscript.exe"

default_pracma = default_pracma_part1 + r"\\R\\win-library\\4.0\\pracma"
default_icon = "C:\\Program Files (x86)\\SharpCap 3.3\\icons\\NP2_ALIGN.ico"

# default_path = r"E:\test"   ##note only one backslash to be compatible with the special folder command outputs
default_path = desktop_folder  ## pick one


#############################################################################################################################
#### default parameters like location
#### change to your liking

default_LAT = 48.084
default_LON = 11.822

default_FOV = 0.88   ## in Deg

default_gain = 100
default_exp = 1

default_sleep = 1   ## in Seconds time for mount to settle

#############################################################################################################################
### initializations

AZ_CORR_DEG = 0.0
ALT_CORR_DEG = 0.0

aborted = False
readyness = False
count = 0

#######################
##### Loop to set ASCOM
def pos_loop(
        # received varaibles from FORM
        LAT= default_LAT,
        LON= default_LON,
        Output=default_path,
        dialog=None
    ):

    def log(text):
        if dialog:
            dialog.process.log(time.strftime("%H%M%S") + " - " + text)
            archivelog = Output + "\\logs\\log_" + time.strftime("%Y%m%d") + ".txt" 
            logfile = open(archivelog, "a")
            logfile.write(time.strftime("%H%M%S") + " - " + text + "\n")
            logfile.close()
        else:
            print text

    def log_status(text):
        if dialog:
            dialog.process.log_status(text)
            archivelog = Output + "\\logs\\log_" + time.strftime("%Y%m%d") + ".txt" 
            logfile = open(archivelog, "a")
            logfile.write(time.strftime("%H%M%S") + " - " + text + "\n")
            logfile.close()
        else:
            print text

    def log_error(text):
        if dialog:
            dialog.paint_alert()
            dialog.process.log_error(time.strftime("%H%M%S") + " - " + text)
            archivelog = Output + "\\logs\\log_" + time.strftime("%Y%m%d") + ".txt" 
            logfile = open(archivelog, "a")
            logfile.write(time.strftime("%H%M%S") + " - " + text + "\n")
            logfile.close()
        else:
            print text

    try:
        # If the site position on the ASCOM sim telescope is correct, then we can get SIDEREALTIME
        # without having to connect with the real telescope, as we would have to specify which one it is.
        # But everybody has the Simulator ... so :)
        # Connect to ASCOM Simulator  
        t = Telescope("ASCOM.Simulator.Telescope")
        
        # write LAT and LON postion to ASCOM site
        t.SiteLatitude = LAT
        t.SiteLongitude = LON
        
        #check overall readyness
        global readyness
        readyness = False
        
        count_ready = 0   

        part1 = str(Output)
        part1 = part1.replace('\\', '\\\\')
        log(str(part1))

        part2 = r"\\star_map.csv"
        starpath = part1 + part2        
  
        if os.path.isfile(default_astap):
            count_ready += 1
            log("ASTAP found")
        else:
            log("ASTAP not found")
        
        
        if os.path.isfile(default_Rscript):
            count_ready += 1
            log("Rscript found")
        else:
            log("Rscript not found")

        
        if os.path.isdir(default_pracma):
            count_ready += 1
            log("Pracma found")
        else:
            log("Pracma not found")
            
        
        if os.path.isfile(starpath):
            count_ready += 1
            log("Star Table found")
        else:
            log("Star Table  not found")
        
        if count_ready == 4:
            readyness = True

    except:
        import traceback
        log_error(traceback.format_exc())

    finally:
        log("ASCOM Updated")
        log("readyness checked" + str(readyness))
        if dialog:
            dialog.pos_end()


#######################
##### Loop to slew mount and take pics
def slew_loop(
        # received varaibles from FORM
        Lat = default_LAT,
        StartALT = 15,
        StartAZ = 80,
        EndAZ = 175,
        no_of_pics = 10,
        Expo = default_exp,
        Gain = default_gain,
        testme = "OFF",
        Output = default_path,            
        dialog = None
    ):

  
    def log(text):
        if dialog:
            dialog.process.log(time.strftime("%H%M%S") + " - " + text)
            archivelog = Output + "\\logs\\log_" + time.strftime("%Y%m%d") + ".txt" 
            logfile = open(archivelog, "a")
            logfile.write(time.strftime("%H%M%S") + " - " + text + "\n")
            logfile.close()
        else:
            print text

    def log_status(text):
        if dialog:
            dialog.process.log_status(text)
            archivelog = Output + "\\logs\\log_" + time.strftime("%Y%m%d") + ".txt" 
            logfile = open(archivelog, "a")
            logfile.write(time.strftime("%H%M%S") + " - " + text + "\n")
            logfile.close()
        else:
            print text

    def log_error(text):
        if dialog:
            dialog.paint_alert()
            dialog.process.log_error(time.strftime("%H%M%S") + " - " + text)
            archivelog = Output + "\\logs\\log_" + time.strftime("%Y%m%d") + ".txt" 
            logfile = open(archivelog, "a")
            logfile.write(time.strftime("%H%M%S") + " - " + text + "\n")
            logfile.close()
        else:
            print text

    try:
        # Prepare camera
        # Set file format to store in
        SharpCap.SelectedCamera.Controls.OutputFormat.Value = "FITS files (*.fits)"
        
        # Set Capture folder to selected path from FORM
        #SharpCap.CaptureFolder = os.path.join(Output)
        
        # Set Expostion and Gain for Camera from FORM
        SharpCap.SelectedCamera.Controls.Exposure.Value = Expo
        #SharpCap.SelectedCamera.Controls.Gain.Value = Gain
        
        # Stop tracking until we give a commmand to go somewhere
        SharpCap.Mounts.SelectedMount.Tracking = False

        # Fetch the SIDEREALTIME from the Simulator Telescope
        t = Telescope("ASCOM.Simulator.Telescope")
        SIDEREALTIME = t.SiderealTime

        # Starting and End Point
        # Geographical Location
        AZ = math.radians(StartAZ)
        AZ_E = math.radians(EndAZ)
        ALT = math.radians(StartALT)
        geoBr = math.radians(Lat)
        sin_geoBr = math.sin(geoBr)
        cos_geoBr = math.cos(geoBr)
        
        # Current position
        # from ASCOM
        RA_NOW = SharpCap.Mounts.SelectedMount.RA
        DEC_NOW = SharpCap.Mounts.SelectedMount.Dec

        
        # Print current position before first move
        log("Mount currently at: RA: %f DEC: %F" %(RA_NOW, DEC_NOW))

        # Define Starting Point as Alt Az Slew
        # careful DEC in degress RA in hours
        # but math in radians
        DEC_START = math.asin( math.sin(ALT) * sin_geoBr - math.cos(ALT) * cos_geoBr * math.cos(AZ + math.pi))
        DEC_START_DEG = math.degrees(DEC_START)

        # Circumvent division through 0 with if loop
        if AZ == 0.:
            AZ = 0.001
        
        # Check in which quadrant we are.
        # if we are in SE & SW, we need to apply a different correction than in NE & NW
        TEST = math.acos((math.sin(ALT) - sin_geoBr * math.sin(DEC_START)) / (cos_geoBr * math.cos(DEC_START))) - (math.pi / 2)
        QUAD_COR = math.copysign(1, TEST)
        
        # Calculate preliminary value of Hour Angle
        A_START = math.atan(math.sin(AZ + math.pi) / ((math.cos(AZ + math.pi) * sin_geoBr) + (math.tan(ALT) * cos_geoBr)))
        
        if math.sin(QUAD_COR) > 0:
            HOUR_ANGLE_START = (math.pi + A_START) % (2* math.pi)
        else:
            HOUR_ANGLE_START = (A_START) % (2* math.pi)
        
        log("HA start: " + str(math.degrees(HOUR_ANGLE_START)))
        
        RA_START = SIDEREALTIME - (math.degrees(HOUR_ANGLE_START) / 15.)
        RA_START = RA_START % 24.        

        log("RA start: " + str(RA_START * 15.))
  
        # Define End Point as Alt Az Slew
        #DEC_END = math.asin( math.sin(ALT) * sin_geoBr + math.cos(ALT) * cos_geoBr * math.cos(AZ_E))
        # We don't vary the DEC Angle
        DEC_END = DEC_START
        DEC_END_DEG = DEC_START_DEG
        
        # Circumvent division through 0 with if loop
        #if AZ_E == 0.:
        #    AZ_E = 0.001
        
        # Check in which quadrant we are.
        # if we are in SE & SW, we need to apply a different correction than in NE & NW
        #TEST = math.acos((math.sin(ALT) - sin_geoBr * math.sin(DEC_END)) / (cos_geoBr * math.cos(DEC_END))) - (math.pi / 2)
        #QUAD_COR = math.copysign(1, TEST)
        
        # Calculate preliminary value of Hour Angle
        A_END = math.atan(math.sin(AZ_E + math.pi) / ((math.cos(AZ_E + math.pi) * sin_geoBr) + (math.tan(ALT) * cos_geoBr)))
        
        # we'll assume for now that we are in the SE or SW quads for now
        #if math.sin(QUAD_COR) > 0:
        #    HOUR_ANGLE_END = (math.pi + A_END) % (2* math.pi)
        #else:
        HOUR_ANGLE_END = (A_END) % (2* math.pi)
        
        log("HA end: " + str(math.degrees(HOUR_ANGLE_END)))        
        
        RA_END = SIDEREALTIME - (math.degrees(HOUR_ANGLE_END) / 15.)
        RA_END = RA_END % 24.  
                       
       
        
        # figure out how many increments are possible until meridian flip
        if RA_END > RA_START :
            DELTA_RA = (RA_START - (RA_END - 24.)) 
        else :
            DELTA_RA = (RA_START - (RA_END ))
        
        # How far to slew for each frame in R.A
        # avoid division by zero
        if no_of_pics > 1:
            RA_SLEW_HOURS = DELTA_RA / (no_of_pics - 1)
        else:
            RA_SLEW_HOURS = DELTA_RA
            
        range_ra = int(no_of_pics)
        
        log("SIDEREALTIME: %f " % SIDEREALTIME)
        log("HA start: %f HA end: %f " % (math.degrees(HOUR_ANGLE_START)/15., math.degrees(HOUR_ANGLE_END)/15.))
        log("RA_START: %f DEC_START: %f " % (RA_START, DEC_START_DEG))
        log("RA_END: %f DEC_END: %f " % (RA_END, DEC_END_DEG))
        log("STEPS: %s" % str(range_ra))
        
        # if the starting ALT is too large, then we exceed the zenith of the azimuth coordinate system
        # which can induce an unwanted meridian flip
        
        global aborted
        
        if (DEC_START_DEG + 5.) > Lat:
            aborted = True
            log("Starting ALT too high")
            log("Risk meridian flip")

        # turn on tracking
        # if we're not tracking, we cannot slew.
        SharpCap.Mounts.SelectedMount.Tracking = True

        RA = RA_START
        DEC = DEC_START_DEG

        # Slew to starting postion
        SharpCap.Mounts.SelectedMount.SlewTo(RA_START, DEC_START)

        # Main FOR loop to
        #   -slew mount to positions
        #   -take pictures
        #   -we'll platesolve and do the math later
        
        # Setup list to store siderealtime
        SIDE = [None] * (range_ra)
        
        # Make count a global variable so we can access it in main FORM
        global count
        count = 0
        count = range_ra

        for ra_l in range(0, (range_ra)):
            if not aborted:
        
                # calculate RA target based on increment of FOR loop
                RA = RA_START - (ra_l * RA_SLEW_HOURS)
                RA = RA % 24.

                # Slew to target
                log("Slewing to: %f, %f" % (RA, DEC))
                SharpCap.Mounts.SelectedMount.SlewTo(RA, DEC)
    
                # Wait for mount to reach target and settle down
                log("Waiting....")
                 
                #wait for the mount to stop slewing
                while SharpCap.Mounts.SelectedMount.Slewing == True:
                    Application.DoEvents()
                    time.sleep(1.)
                
                for i in range(0, int(default_sleep)):
                    Application.DoEvents()
                    time.sleep(1.)
        
                log("CAPTURING IMAGE.")
        
                #store sidereal time
                SIDE[ra_l] = t.SiderealTime
                
                part1 = str(Output)
                part1 = part1.replace('\\', '\\\\')
                log(str(part1))
                
                part2 = r"\\pictures"
                
                part3 = r"\\%03d.fits" % ra_l
                
                path = part1 + part2 + part3
                                
                SharpCap.SelectedCamera.CaptureSingleFrameTo(path)
        
        # Wait for Plate Solve to finish up
        time.sleep(3.)

    except:
        import traceback
        log_error(traceback.format_exc())

    finally:
        if dialog:
            dialog.slew_end()
            dialog.slew_finish()
        log("End Capture")


#######################
##### Loop to send pictures off to plate solving
def solve_loop(
        LAT = default_LAT,
        LON = default_LON,
        PRESSURE = 1013,
        AMB_TEMP = 5,
        FOV = default_FOV,
        testme = "OFF",
        Output = default_path,            
        dialog = None
    ):

   
    def log(text):
        if dialog:
            dialog.process.log(time.strftime("%H%M%S") + " - " + text)
            archivelog = Output + "\\logs\\log_" + time.strftime("%Y%m%d") + ".txt" 
            logfile = open(archivelog, "a")
            logfile.write(time.strftime("%H%M%S") + " - " + text + "\n")
            logfile.close()
        else:
            print text

    def log_status(text):
        if dialog:
            dialog.process.log_status(text)
            archivelog = Output + "\\logs\\log_" + time.strftime("%Y%m%d") + ".txt" 
            logfile = open(archivelog, "a")
            logfile.write(time.strftime("%H%M%S") + " - " + text + "\n")
            logfile.close()
        else:
            print text

    def log_error(text):
        if dialog:
            dialog.paint_alert()
            dialog.process.log_error(time.strftime("%H%M%S") + " - " + text)
            archivelog = Output + "\\logs\\log_" + time.strftime("%Y%m%d") + ".txt" 
            logfile = open(archivelog, "a")
            logfile.write(time.strftime("%H%M%S") + " - " + text + "\n")
            logfile.close()
        else:
            print text

    try:
        # Setup command to launch ASTAP and solve our pictures   
        # results are stored in an .ini file
        # 
        a = default_astap
        b = "-f"
        
        # Check if test box is selected and alter path to look for pictures
        # based on that
        part1 = str(Output)
        part1 = part1.replace('\\', '\\\\')
        log(str(part1))   

        base = part1 + r"\\"
        part2 = r"\\pictures\\"           
        
        log(str(testme))
        
        if testme == 1:
            part2 = r"\\test1\\"
            FOV = 0.88
            
        if testme == 2:
            part2 = r"\\test2\\"   
            FOV = 0.49
            
        if testme == 3:
            part2 = r"\\test3\\"                            
            FOV = 0.49
     
        if testme == 4:
            part2 = r"\\test4\\"   
            FOV = 0.49

        c = part1 + part2
            
        d = "-fov %.2f -z 2" % FOV

        # we need the path once with backslashes at the end
        # and once without (which is here)
        path = c[:-2]
        
        # Read files in path to find fits files (the pictures)
        dirListing = os.listdir(path)
        editFiles = []
        for item in dirListing:
            if item.endswith('.fits'):
                editFiles.append(item)
        log(str(editFiles))
        global count, white_ball, color_ball
        count = len(editFiles)
        catch = 0
        white_ball = 0
        color_ball = []
        for i in range(0, count):
            color_ball.append(0)
        
        
        while catch < count:
            if not aborted:
                # Launch ASTAP with normal parameters
            
                e = c + str(editFiles[catch])
                mycmd="%s %s %s %s" %(a,b,e,d)
                log(mycmd)
                my = subprocess.Popen(shlex.split(mycmd), stdout=subprocess.PIPE)
                out, err = my.communicate()
                exit_code = my.wait()
                log(str(my.returncode))
                if my.returncode == 0:
                    white_ball += 1
                    color_ball[catch] = 1
            
                # If ASTAP did not solve on first go around...
                # re-start ASTAP with another downsampling setting
                if my.returncode == 1:
                    f = "-fov %.2f -z 3" % FOV
                    mzcmd="%s %s %s %s" %(a,b,e,f)
                    mz = subprocess.Popen(shlex.split(mzcmd), stdout=subprocess.PIPE)
                    out, err = mz.communicate()
                    exit_code = mz.wait()
                    log(str(mz.returncode))
                    if mz.returncode == 0:
                        white_ball += 1
                        color_ball[catch] = 1
                
                    # If ASTAP did not solve on first go around...
                    # re-start ASTAP with another downsampling setting
                    if mz.returncode == 1:
                        g = "-fov %.2f -z 4" % FOV
                        mxcmd="%s %s %s %s" %(a,b,e,g)
                        mx = subprocess.Popen(shlex.split(mxcmd), stdout=subprocess.PIPE)
                        out, err = mx.communicate()
                        exit_code = mx.wait()
                        log(str(mx.returncode))                    
                        if mx.returncode == 0:
                            white_ball += 1
                            color_ball[catch] = 1
                        
                catch += 1
        
        # Re-read files in path to check for results of plate solving
        # filter for result files (.ini)        
        dirListing = os.listdir(path)          
        editFiles = []
        for item in dirListing:
            if item.endswith('.ini'):
                editFiles.append(item)
        
        countps = len(editFiles)        
        catchps = 0

        cols = 6
        global arr
        arr = [[0 for i in range(cols)] for j in range(countps)]
        
        while catchps < countps:
            if not aborted:
                pspath = c + str(editFiles[catchps])
                ini_file = open(pspath)
                ini_lines = ini_file.readlines()
            
                # check if more than 6 lines are stored in .ini
                # if there are less, then the solve was negative
                if len(ini_lines) > 6:
            
                    # Grab lines 3 and 4, which contain the RA and DEC solutions for
                    # the center of the picture
                    log(str(ini_lines[3]))
                    log(str(ini_lines[4]))
                
                    RA_raw = str(ini_lines[3])
                    DEC_raw = str(ini_lines[4])
    
                    # Split line at the = character
                    lhs_RA, rhs_RA = RA_raw.split('=')
                    lhs_DEC, rhs_DEC = DEC_raw.split('=')
                
                    # convert to float so we have a number and not a sting
                    # also float can deal with scientific notation like 1.234E-04
                    RA_ps = float(rhs_RA)
                    DEC_ps = float(rhs_DEC)
            
                else:
                    RA_ps = 0
                    DEC_ps = 0
                
                ini_file.close()
            
                # Store values to an array (or 2d list), which we will later write to a file
                arr[catchps][0] = catchps
                arr[catchps][1] = RA_ps
                arr[catchps][2] = DEC_ps
            
                catchps += 1
            
        # read sharpcap .txt file, which was created for each picture
        # here we'll grab the time the picture was taken to convert to JD, and HA
        editFiles = []
        for item in dirListing:
            if item.endswith('.txt'):
                editFiles.append(item)
        
        countps = len(editFiles)        
        catchps = 0

        while catchps < countps:
            if not aborted:
                pspath = c + str(editFiles[catchps])
                ini_file = open(pspath)
                ini_lines = ini_file.readlines()
            
                useful = []
                for line in ini_lines:
                    if "TimeStamp=" in line:
                        break
                log(str(line))
            
                Time_raw = str(line)

                lhs_Time, rhs_Time = Time_raw.split('=')

                Time_year, Time_month, Rest = rhs_Time.split('-')
                Time_day, Time_str = Rest.split('T')
            
                Time_hour_raw, Time_min_raw, Time_sec_raw = Time_str.split(':')
                Time_hour = float(Time_hour_raw)
                Time_min = float(Time_min_raw)
                Time_sec_almost = Time_sec_raw[:-2]
                Time_sec = float(Time_sec_almost)
            
                # Julian Day
                JD = int(365.25 * (int(Time_year) + 4716)) + int(30.6001 * (int(Time_month) + 1)) + int(Time_day) - 13 - 1524.5 + (Time_hour / 24) + (Time_min / 1440) + (Time_sec / 86400)
            
                T = ((JD - 2451545.0) / 36525)
                theta0 = 280.46061837 + 360.98564736629 * (JD - 2451545.0) + (0.000387933 * T * T) - (T * T * T / 38710000.0)
            
                # Greenwich Sidereal Time in degrees
                gmstRA = theta0 % 360
            
                # Local Sidereal Time in degrees
                LST = gmstRA + LON
            
                LST = LST % 360
            
                # Hour Angle is LST - RA
                HA = LST - arr[catchps][1]
                
                ini_file.close()
                # store further values to array
                arr[catchps][3] = JD
                arr[catchps][4] = LST
                arr[catchps][5] = HA
                
                log(str(theta0))
                log(str(gmstRA))
                log(str(LON))
                log(str(LST))
                log(str(HA))

                catchps += 1
        
        # now we'll write the array together with the site data to a csv file, which we'll then read with Rscript
        header1 = "LAT,LON,PRESSURE,AMB_TEMP"
        header2 = "NR,RA,DEC,JD,LST,HA"
        comb_string = str(LAT) + "," + str(LON) + "," + str(PRESSURE) + "," + str(AMB_TEMP)
        
        if not aborted:
            with open(base + "ps_positions.csv", "w") as myfile:
                myfile.write("%s\n" % header1)
                myfile.write("%s\n" % comb_string)
                myfile.write("%s\n" % header2)
       
                for row in arr:
                    myfile.write("%s\n" % ','.join(str(col) for col in row))      

    except:
        import traceback
        log_error(traceback.format_exc())

    finally:
        if dialog:
            dialog.slew_end()
            dialog.paint_finish()
        log("End Capture")
        log(str(white_ball) + " of" + str(count) + " platesolved!")


#######################
##### Loop to send plate solving results to Rscript to calculate correction angles
def analysis_loop(
        # received varaibles from FORM
        LAT = default_LAT,
        LON = default_LON,
        Output = default_path,
        dialog = None
    ):

 
    def log(text):
        if dialog:
            dialog.process.log(time.strftime("%H%M%S") + " - " + text)
            archivelog = Output + "\\logs\\log_" + time.strftime("%Y%m%d") + ".txt" 
            logfile = open(archivelog, "a")
            logfile.write(time.strftime("%H%M%S") + " - " + text + "\n")
            logfile.close()
        else:
            print text

    def log_status(text):
        if dialog:
            dialog.process.log_status(text)
            archivelog = Output + "\\logs\\log_" + time.strftime("%Y%m%d") + ".txt" 
            logfile = open(archivelog, "a")
            logfile.write(time.strftime("%H%M%S") + " - " + text + "\n")
            logfile.close()
        else:
            print text

    def log_error(text):
        if dialog:
            dialog.paint_alert()
            dialog.process.log_error(time.strftime("%H%M%S") + " - " + text)
            archivelog = Output + "\\logs\\log_" + time.strftime("%Y%m%d") + ".txt" 
            logfile = open(archivelog, "a")
            logfile.write(time.strftime("%H%M%S") + " - " + text + "\n")
            logfile.close()
        else:
            print text

    try:
        # call RScript in the installation path of R
        # then execute our script, which will provide the correction angles
        
        part1 = str(Output)
        part1 = part1.replace('\\', '\\\\')
        log(str(part1))

        part2 = r"\\scripts\\astro_v32.R"
        c = part1 + part2
        
        
        arg_part1 = str(Output)
        arg_part1 = arg_part1.replace('\\', '/')
        
        arg_part2 = r"/ps_positions.csv"
        arg_part3 = r"/ps_corrections.csv"
        
        arg1 = arg_part1 + arg_part2
        arg2 = arg_part1 + arg_part3
        
        a = default_Rscript

        mycmd="%s --vanilla %s %s %s" % (a,c,arg1,arg2)
        log(mycmd)

        my = subprocess.Popen(shlex.split(mycmd), shell = True, stdout=subprocess.PIPE)
        out, err = my.communicate()
        exit_code = my.wait()
        log(str(my.returncode))
        
        global AZ_CORR_DEG, ALT_CORR_DEG
        
        if my.returncode == 0:
        
            part1 = str(Output)
            part1 = part1.replace('\\', '\\\\')
            log(str(part1))    

            base = part1 + r"\\"  
                
            pspath = base + "ps_corrections.csv"
            
            ini_file = open(pspath)
            ini_lines = ini_file.readlines()        

            raw = str(ini_lines[1])
            AZ_CORR_DEG_raw, ALT_CORR_DEG_raw = raw.split(',')
            AZ_CORR_DEG = float(AZ_CORR_DEG_raw)
            ALT_CORR_DEG = float(ALT_CORR_DEG_raw)
            ini_file.close()        

    except:
        import traceback
        log_error(traceback.format_exc())

    finally:
        log("Correction Updated")
        if dialog:
            dialog.analysis_finish()


#######################
##### Loop slew to first point
##### lookup bright star from table, go there, sync mount, set focusers up

def SEalign_loop(
        # received varaibles from FORM
        LAT= default_LAT,
        LON= default_LON,
        Output=desktop_folder,
        dialog=None
    ):

    def log(text):
        if dialog:
            dialog.process.log(time.strftime("%H%M%S") + " - " + text)
            archivelog = Output + "\\logs\\log_" + time.strftime("%Y%m%d") + ".txt" 
            logfile = open(archivelog, "a")
            logfile.write(time.strftime("%H%M%S") + " - " + text + "\n")
            logfile.close()
        else:
            print text

    def log_status(text):
        if dialog:
            dialog.process.log_status(text)
            archivelog = Output + "\\logs\\log_" + time.strftime("%Y%m%d") + ".txt" 
            logfile = open(archivelog, "a")
            logfile.write(time.strftime("%H%M%S") + " - " + text + "\n")
            logfile.close()
        else:
            print text

    def log_error(text):
        if dialog:
            dialog.paint_alert()
            dialog.process.log_error(time.strftime("%H%M%S") + " - " + text)
            archivelog = Output + "\\logs\\log_" + time.strftime("%Y%m%d") + ".txt" 
            logfile = open(archivelog, "a")
            logfile.write(time.strftime("%H%M%S") + " - " + text + "\n")
            logfile.close()
        else:
            print text

    try:
        # If the site position on the ASCOM sim telescope is correct, then we can get SIDEREALTIME
        # without having to connect with the real telescope, as we would have to specify which one it is.
        # But everybody has the Simulator ... so :)
        # Connect to ASCOM Simulator  
        t = Telescope("ASCOM.Simulator.Telescope")
        SIDEREALTIME = t.SiderealTime
        
        ### SE Azimuth
        SE_AZ = math.radians(135.)
        HOUR_ANGLE = 21.5 / 24. * 2 * math.pi   ### poor mans approach
        
        RA_TARGET = SIDEREALTIME - (math.degrees(HOUR_ANGLE) / 15.)
        RA_TARGET = RA_TARGET % 24
        RA_LOOK = round((RA_TARGET * 3), 0) * 5
        RA_LOOK = RA_LOOK % 360
        
        part1 = str(Output)
        part1 = part1.replace('\\', '\\\\')
        log(str(part1))

        part2 = r"\\star_map.csv"
        starpath = part1 + part2
                
        data = list(csv.reader(open(starpath)))

        length = len(data)
        log("# of bright star records found")
        log(str(length))

        STAR = "none"
        RA_GO_raw = 0
        DEC_GO_raw = 0
        for i in range(1 , length):
            if int(data[i][5]) == int(RA_LOOK):
                RA_GO_raw = data[i][1]
                DEC_GO_raw = data[i][2]
                STAR = data[i][0]                
                break
        
        log("going to...")
        log(str(STAR))
        log("at coords:")
        log(str(RA_GO_raw))
        log(str(DEC_GO_raw))
        
        RA_GO = float(RA_GO_raw) / 15.
        DEC_GO = float(DEC_GO_raw)
        
        if int(RA_GO) != 0 and int(DEC_GO) != 0:
            log("Mount tracking activated")
            SharpCap.Mounts.SelectedMount.Tracking = True
            # Slew to starting postion
            SharpCap.Mounts.SelectedMount.SlewTo(RA_GO, DEC_GO)
            
            log("waiting for mount to complete slew")
            #wait for the mount to stop slewing
            while SharpCap.Mounts.SelectedMount.Slewing == True:
                #Thread.Sleep(int(1000))
                Application.DoEvents()
                time.sleep(1.)

            log("wait for mount to settle")
            for i in range(0, int(default_sleep)):
                Application.DoEvents()
                time.sleep(1.)
            
            log("shot picture and sync")
            SharpCap.Mounts.SelectedMount.SolveAndSync()
             

    except:
        import traceback
        log_error(traceback.format_exc())

    finally:
        log("SE align performed")
        if dialog:
            dialog.SEalign_end()


#######################
##### Loop slew to manual alignment point
##### lookup bright star from table, go there, sync mount

def Salign_loop(
        # received varaibles from FORM
        LAT = default_LAT,
        LON = default_LON,
        Output = default_path,
        dialog = None
    ):


    def log(text):
        if dialog:
            dialog.process.log(time.strftime("%H%M%S") + " - " + text)
            archivelog = Output + "\\logs\\log_" + time.strftime("%Y%m%d") + ".txt" 
            logfile = open(archivelog, "a")
            logfile.write(time.strftime("%H%M%S") + " - " + text + "\n")
            logfile.close()
        else:
            print text

    def log_status(text):
        if dialog:
            dialog.process.log_status(text)
            archivelog = Output + "\\logs\\log_" + time.strftime("%Y%m%d") + ".txt" 
            logfile = open(archivelog, "a")
            logfile.write(time.strftime("%H%M%S") + " - " + text + "\n")
            logfile.close()
        else:
            print text

    def log_error(text):
        if dialog:
            dialog.paint_alert()
            dialog.process.log_error(time.strftime("%H%M%S") + " - " + text)
            archivelog = Output + "\\logs\\log_" + time.strftime("%Y%m%d") + ".txt" 
            logfile = open(archivelog, "a")
            logfile.write(time.strftime("%H%M%S") + " - " + text + "\n")
            logfile.close()
        else:
            print text

    try:
        # If the site position on the ASCOM sim telescope is correct, then we can get SIDEREALTIME
        # without having to connect with the real telescope, as we would have to specify which one it is.
        # But everybody has the Simulator ... so :)
        # Connect to ASCOM Simulator  
        t = Telescope("ASCOM.Simulator.Telescope")
        SIDEREALTIME = t.SiderealTime
        
        ### S Azimuth
        S_AZ = math.radians(172.5)
        HOUR_ANGLE = math.atan(math.tan(S_AZ) / math.sin(math.radians(LAT)))
        #HOUR_ANGLE = math.radians(360.)
        
        RA_TARGET = SIDEREALTIME - (math.degrees(HOUR_ANGLE) / 15.)
        RA_TARGET = RA_TARGET % 24
        RA_LOOK = round((RA_TARGET * 3), 0) * 5 - 5 
        RA_LOOK = RA_LOOK % 360
        
        part1 = str(Output)
        part1 = part1.replace('\\', '\\\\')
        log(str(part1))

        part2 = r"\\star_map.csv"
        starpath = part1 + part2
                
        data = list(csv.reader(open(starpath)))

        length = len(data)
        log("# of bright star records found")
        log(str(length))

        STAR = "none"
        RA_GO_raw = 0
        DEC_GO_raw = 0
        for i in range(1 , length):
            if int(data[i][5]) == int(RA_LOOK):
                RA_GO_raw = data[i][1]
                DEC_GO_raw = data[i][2]
                STAR = data[i][0]                
                break
        
        log("going to...")
        log(str(STAR))
        log("at coords:")
        log(str(RA_GO_raw))
        log(str(DEC_GO_raw))
        
        RA_GO = float(RA_GO_raw) / 15.
        DEC_GO = float(DEC_GO_raw)
        
        if int(RA_GO) != 0 and int(DEC_GO) != 0:
            log("Mount tracking activated")
            SharpCap.Mounts.SelectedMount.Tracking = True
            # Slew to starting postion
            SharpCap.Mounts.SelectedMount.SlewTo(RA_GO, DEC_GO)
            
            log("waiting for mount to complete slew")
            #wait for the mount to stop slewing
            while SharpCap.Mounts.SelectedMount.Slewing == True:
                #Thread.Sleep(int(1000))
                Application.DoEvents()
                time.sleep(1.)

            log("wait for mount to settle")
            #wait for the mount to settle
            for i in range(0, int(default_sleep)):
                Application.DoEvents()
                time.sleep(1.)
            
            log("shot picture and sync")
            SharpCap.Mounts.SelectedMount.SolveAndSync()
             

    except:
        import traceback
        log_error(traceback.format_exc())

    finally:
        log("S align performed")
        if dialog:
            dialog.Salign_end()


#######################
##### Correct for ALT misalignment
##### 

def corrALT_loop(
        # received varaibles from FORM
        LAT = default_LAT,
        LON = default_LON,
        Output = default_path,
        dialog = None
    ):


    def log(text):
        if dialog:
            dialog.process.log(time.strftime("%H%M%S") + " - " + text)
            archivelog = Output + "\\logs\\log_" + time.strftime("%Y%m%d") + ".txt" 
            logfile = open(archivelog, "a")
            logfile.write(time.strftime("%H%M%S") + " - " + text + "\n")
            logfile.close()
        else:
            print text

    def log_status(text):
        if dialog:
            dialog.process.log_status(text)
            archivelog = Output + "\\logs\\log_" + time.strftime("%Y%m%d") + ".txt" 
            logfile = open(archivelog, "a")
            logfile.write(time.strftime("%H%M%S") + " - " + text + "\n")
            logfile.close()
        else:
            print text

    def log_error(text):
        if dialog:
            dialog.paint_alert()
            dialog.process.log_error(time.strftime("%H%M%S") + " - " + text)
            archivelog = Output + "\\logs\\log_" + time.strftime("%Y%m%d") + ".txt" 
            logfile = open(archivelog, "a")
            logfile.write(time.strftime("%H%M%S") + " - " + text + "\n")
            logfile.close()
        else:
            print text

    try:
        # If the site position on the ASCOM sim telescope is correct, then we can get SIDEREALTIME
        # without having to connect with the real telescope, as we would have to specify which one it is.
        # But everybody has the Simulator ... so :)
        # Connect to ASCOM Simulator  
        t = Telescope("ASCOM.Simulator.Telescope")
        SIDEREALTIME = t.SiderealTime

        # input correction angles
        global AZ_CORR_DEG, ALT_CORR_DEG

        # We want to shift the mount in the opposite direction of the misalignment, so the manual
        # correction will take us to where we want to be
        COR_ALT = math.radians(-ALT_CORR_DEG)
        COR_AZ = math.radians(-AZ_CORR_DEG)

        # Geographical Location
        geoBr = math.radians(LAT)
        sin_geoBr = math.sin(geoBr)
        cos_geoBr = math.cos(geoBr)

        # Current position
        # from SharpCap
        RA_NOW_HR = SharpCap.Mounts.SelectedMount.RA
        RA_NOW_DEG = RA_NOW_HR * 180 / 12
        DEC_NOW_DEG = SharpCap.Mounts.SelectedMount.Dec
        DEC_NOW = math.radians(DEC_NOW_DEG)

        # Print current position before first move
        log("Mount currently at: RA: %f DEC: %F" %(RA_NOW_DEG, DEC_NOW_DEG))

        log("transforming from equatorial coordinates to local horizontal coordinates")

        HA_NOW = math.radians((t.SiderealTime - RA_NOW_HR) * 15.)
        HA_NOW_DEG = math.degrees(HA_NOW)

        log("Mount currently at: HA: %f DEC: %F" %(HA_NOW_DEG, DEC_NOW_DEG))


        #sin geo_Br * sin DEC
        SS = sin_geoBr * math.sin(DEC_NOW)
  
        #cos geo_Br * cos DEC
        CC = cos_geoBr * math.cos(DEC_NOW)
  
        #cos geo_Br * tan DEC
        CT = cos_geoBr * math.tan(DEC_NOW)

        #calculate altitude from DEC and sidereal
        ALT_NOW = math.asin(SS + CC * math.cos(HA_NOW))
        ALT_NOW_DEG = math.degrees(ALT_NOW)

        plate_test =  math.sin(HA_NOW) / (sin_geoBr * math.cos(HA_NOW) - CT)
        # correction factor
        # if plate$test < 0 then 2 pi ... else 1 pi
        # sign is either -1 or +1, if you add that to three and divde by two
        # then you get the 2pi or 1pi
        plate_cor1 = - (math.copysign(1, plate_test) - 3. ) / 2 * math.pi
  
        # correction factor 2
        # if plate$side is larger than pi then pi else 0
        plate_cor2 = ((HA_NOW + 2 * math.pi) % (2* math.pi)) > math.pi
        plate_cor2 = int(plate_cor2 == True) * math.pi

        # calculate azimuth from sidereal, DEC and corrections
        AZ_NOW = (plate_cor1 + plate_cor2 + math.atan(math.sin(HA_NOW) / (sin_geoBr * math.cos(HA_NOW) - CT))) % (2 *math.pi)
        AZ_NOW_DEG = math.degrees(AZ_NOW)

        log("Mount currently at: ALT: %f AZ: %F" %(ALT_NOW_DEG, AZ_NOW_DEG))

        log("transforming from to local horizontal coordinates to Cartesian coordinates")

        # calculate cartesian coordinates
        x = math.cos(AZ_NOW) * math.cos(ALT_NOW)
        y = math.sin(AZ_NOW) * math.cos(ALT_NOW)
        z = math.sin(ALT_NOW)

        log("Mount currently at: x: %f y: %F z: %f" %(x, y, z))

        log("Applying rotation matrix for ALT correction")

        # set up rotation matrix for y-axis rotation
        xx = math.cos(-COR_ALT)
        xy = 0.0
        xz = math.sin(-COR_ALT)

        yx = 0.0
        yy = 1.0
        yz = 0.0

        zx = -math.sin(-COR_ALT)
        zy = 0.0
        zz = math.cos(-COR_ALT)

        # calculate rotated coordinates
        x_n1 = x * xx + y * yx + z * zx
        y_n1 = x * xy + y * yy + z * zy
        z_n1 = x * xz + y * yz + z * zz

        log("Mount with Alt rotation at: x_n1: %f y_n1: %F z_n1: %f" %(x_n1, y_n1, z_n1))


        #ALT corrected local horizontal coordinates
        ALT_C1 = math.asin(z_n1)
        AZ_C1 = math.atan(y_n1/x_n1) + math.pi
        ALT_C1_DEG = math.degrees(ALT_C1)
        AZ_C1_DEG = math.degrees(AZ_C1)

        log("Mount with Alt rotation at: ALT: %f AZ: %F" %(ALT_C1_DEG, AZ_C1_DEG))

        log("Applying correction angles, ALT: %f" % ALT_CORR_DEG)

        # Define Starting Point as Alt Az Slew
        # careful DEC in degress RA in hours
        # but math in radians

        DEC_NEW1 = math.asin( math.sin(ALT_C1) * sin_geoBr + math.cos(ALT_C1) * cos_geoBr * math.cos(AZ_C1))
        HA_NEW1 = math.asin(-math.sin(AZ_C1) * math.cos(ALT_C1) / math.cos(DEC_NEW1))
        RA_NEW1_HR = t.SiderealTime  - math.degrees(HA_NEW1) / 15.

        RA_NEW1_DEG = RA_NEW1_HR * 15.
        DEC_NEW1_DEG = math.degrees(DEC_NEW1)

        # Updated position
     
        log("Moving to correction position ALT: RA: %f DEC: %f " % (RA_NEW1_DEG, DEC_NEW1_DEG))

        SharpCap.Mounts.SelectedMount.Tracking = True
        # Slew to starting postion
        SharpCap.Mounts.SelectedMount.SlewTo(RA_NEW1_HR, DEC_NEW1_DEG)
            
        log("waiting for mount to complete slew")
        #wait for the mount to stop slewing
        while SharpCap.Mounts.SelectedMount.Slewing == True:
            Application.DoEvents()
            time.sleep(1.)
      
        log("Now move the target back to center with the manual mount adjustments")

    except:
        import traceback
        log_error(traceback.format_exc())

    finally:
        log("ALT Offset")
        if dialog:
            dialog.corrALT_end()


#######################
##### Correct for ALT misalignment
##### 

def corrAZ_loop(
        # received varaibles from FORM
        LAT = default_LAT,
        LON = default_LON,
        Output = default_path,
        dialog = None
    ):


    def log(text):
        if dialog:
            dialog.process.log(time.strftime("%H%M%S") + " - " + text)
            archivelog = Output + "\\logs\\log_" + time.strftime("%Y%m%d") + ".txt" 
            logfile = open(archivelog, "a")
            logfile.write(time.strftime("%H%M%S") + " - " + text + "\n")
            logfile.close()
        else:
            print text

    def log_status(text):
        if dialog:
            dialog.process.log_status(text)
            archivelog = Output + "\\logs\\log_" + time.strftime("%Y%m%d") + ".txt" 
            logfile = open(archivelog, "a")
            logfile.write(time.strftime("%H%M%S") + " - " + text + "\n")
            logfile.close()
        else:
            print text

    def log_error(text):
        if dialog:
            dialog.paint_alert()
            dialog.process.log_error(time.strftime("%H%M%S") + " - " + text)
            archivelog = Output + "\\logs\\log_" + time.strftime("%Y%m%d") + ".txt" 
            logfile = open(archivelog, "a")
            logfile.write(time.strftime("%H%M%S") + " - " + text + "\n")
            logfile.close()
        else:
            print text

    try:
        # If the site position on the ASCOM sim telescope is correct, then we can get SIDEREALTIME
        # without having to connect with the real telescope, as we would have to specify which one it is.
        # But everybody has the Simulator ... so :)
        # Connect to ASCOM Simulator  
        t = Telescope("ASCOM.Simulator.Telescope")
        SIDEREALTIME = t.SiderealTime

        # input correction angles
        global AZ_CORR_DEG, ALT_CORR_DEG

        # We want to shift the mount in the opposite direction of the misalignment, so the manual
        # correction will take us to where we want to be
        COR_ALT = math.radians(-ALT_CORR_DEG)
        COR_AZ = math.radians(-AZ_CORR_DEG)

        # Geographical Location
        geoBr = math.radians(LAT)
        sin_geoBr = math.sin(geoBr)
        cos_geoBr = math.cos(geoBr)

        # Current position
        # from SharpCap
        RA_NOW_HR = SharpCap.Mounts.SelectedMount.RA
        RA_NOW_DEG = RA_NOW_HR * 180 / 12
        DEC_NOW_DEG = SharpCap.Mounts.SelectedMount.Dec
        DEC_NOW = math.radians(DEC_NOW_DEG)

        # Print current position before first move
        log("Mount currently at: RA: %f DEC: %F" %(RA_NOW_DEG, DEC_NOW_DEG))

        log("transforming from equatorial coordinates to local horizontal coordinates")

        HA_NOW = math.radians((t.SiderealTime - RA_NOW_HR) * 15.)
        HA_NOW_DEG = math.degrees(HA_NOW)

        log("Mount currently at: HA: %f DEC: %F" %(HA_NOW_DEG, DEC_NOW_DEG))


        #sin geo_Br * sin DEC
        SS = sin_geoBr * math.sin(DEC_NOW)
  
        #cos geo_Br * cos DEC
        CC = cos_geoBr * math.cos(DEC_NOW)
  
        #cos geo_Br * tan DEC
        CT = cos_geoBr * math.tan(DEC_NOW)

        #calculate altitude from DEC and sidereal
        ALT_NOW = math.asin(SS + CC * math.cos(HA_NOW))
        ALT_NOW_DEG = math.degrees(ALT_NOW)

        plate_test =  math.sin(HA_NOW) / (sin_geoBr * math.cos(HA_NOW) - CT)
        # correction factor
        # if plate$test < 0 then 2 pi ... else 1 pi
        # sign is either -1 or +1, if you add that to three and divde by two
        # then you get the 2pi or 1pi
        plate_cor1 = - (math.copysign(1, plate_test) - 3. ) / 2 * math.pi
  
        # correction factor 2
        # if plate$side is larger than pi then pi else 0
        plate_cor2 = ((HA_NOW + 2 * math.pi) % (2* math.pi)) > math.pi
        plate_cor2 = int(plate_cor2 == True) * math.pi

        # calculate azimuth from sidereal, DEC and corrections
        AZ_NOW = (plate_cor1 + plate_cor2 + math.atan(math.sin(HA_NOW) / (sin_geoBr * math.cos(HA_NOW) - CT))) % (2 *math.pi)
        AZ_NOW_DEG = math.degrees(AZ_NOW)

        log("Mount currently at: ALT: %f AZ: %F" %(ALT_NOW_DEG, AZ_NOW_DEG))

        log("transforming from to local horizontal coordinates to Cartesian coordinates")

        # calculate cartesian coordinates
        x = math.cos(AZ_NOW) * math.cos(ALT_NOW)
        y = math.sin(AZ_NOW) * math.cos(ALT_NOW)
        z = math.sin(ALT_NOW)

        log("Mount currently at: x: %f y: %F z: %f" %(x, y, z))

        log("Applying rotation matrix for ALT correction")

        # set up rotation matrix for y-axis rotation
        xx = math.cos(-COR_ALT)
        xy = 0.0
        xz = math.sin(-COR_ALT)

        yx = 0.0
        yy = 1.0
        yz = 0.0

        zx = -math.sin(-COR_ALT)
        zy = 0.0
        zz = math.cos(-COR_ALT)

        # calculate rotated coordinates
        x_n1 = x * xx + y * yx + z * zx
        y_n1 = x * xy + y * yy + z * zy
        z_n1 = x * xz + y * yz + z * zz

        log("Mount with Alt rotation at: x_n1: %f y_n1: %F z_n1: %f" %(x_n1, y_n1, z_n1))

        # set up rotation matrix for z-axis rotation
        xx = math.cos(-COR_AZ)
        xy = -math.sin(-COR_AZ)
        xz = 0.0

        yx = math.sin(-COR_AZ)
        yy = math.cos(-COR_AZ)
        yz = 0.0

        zx = 0.0
        zy = 0.0
        zz = 1.0

        # calculate rotated coordinates
        x_n2 = x_n1 * xx + y_n1 * yx + z_n1 * zx
        y_n2 = x_n1 * xy + y_n1 * yy + z_n1 * zy
        z_n2 = x_n1 * xz + y_n1 * yz + z_n1 * zz

        log("Mount with Alt and Az rotation at: x_n2: %f y_n2: %F z_n2: %f" %(x_n2, y_n2, z_n2))

        #ALT and AZ corrected local horizontal coordinates
        ALT_C2 = math.asin(z_n2)
        AZ_C2 = math.atan(y_n2/x_n2) + math.pi
        ALT_C2_DEG = math.degrees(ALT_C2)
        AZ_C2_DEG = math.degrees(AZ_C2)

        log("Mount with Alt and AZ rotation at: ALT: %f AZ: %F" %(ALT_C2_DEG, AZ_C2_DEG))

        log("Applying correction angles, ALT: %f and AZ: %f" % (ALT_CORR_DEG, AZ_CORR_DEG))

        # Updated position
        DEC_NEW2 = math.asin( math.sin(ALT_C2) * sin_geoBr + math.cos(ALT_C2) * cos_geoBr * math.cos(AZ_C2))
        HA_NEW2 = math.asin(-math.sin(AZ_C2) * math.cos(ALT_C2) / math.cos(DEC_NEW2))
        RA_NEW2_HR = t.SiderealTime  - math.degrees(HA_NEW2) / 15.

        RA_NEW2_DEG = RA_NEW2_HR * 15.
        DEC_NEW2_DEG = math.degrees(DEC_NEW2)

        log("Moving to correction position ALT +AZ: RA: %f DEC: %f " % (RA_NEW2_DEG, DEC_NEW2_DEG))

        SharpCap.Mounts.SelectedMount.Tracking = True
        # Slew to starting postion
        SharpCap.Mounts.SelectedMount.SlewTo(RA_NEW2_HR, DEC_NEW2_DEG)
            
        log("waiting for mount to complete slew")
        #wait for the mount to stop slewing
        while SharpCap.Mounts.SelectedMount.Slewing == True:
            #Thread.Sleep(int(1000))
            Application.DoEvents()
            time.sleep(1.)

        log("Now move the target back to center with the manual mount adjustments")  

    except:
        import traceback
        log_error(traceback.format_exc())

    finally:
        log("ALT and AZ Offset")
        if dialog:
            dialog.corrAZ_end()



# definition of drop down field
class ComboInput(Panel):

    def __init__(self, label, default_value="OFF"):
        super(ComboInput, self).__init__()
        self.Height = 30
        self.default_value = default_value

        label_widget = Label()
        label_widget.Text = label +":"
        label_widget.Dock = DockStyle.Right
        label_widget.AutoSize = True
        label_widget.Padding = Padding(2)

        input_widget = ComboBox()
        input_widget.Items.Add("OFF")
        input_widget.Items.Add("Test1")
        input_widget.Items.Add("Test2")
        input_widget.Items.Add("Test3")
        input_widget.Items.Add("Test4")
        input_widget.SelectedItem  = str(default_value)
        input_widget.Dock = DockStyle.Right

        self.Controls.Add(label_widget)
        self.Controls.Add(input_widget)
        self.label = label_widget
        self.input = input_widget

    def resize_to_contents(self):
        self.Width = self.label.Width + self.input.Width

    def get_label(self):
        return self.label.Text[:-1]

    def get_value(self):
        return self.input.SelectedIndex        



# definition of value field        
class ValueInput(Panel):

    def __init__(self, label, default_value=""):
        super(ValueInput, self).__init__()
        self.Height = 30
        self.default_value = default_value

        label_widget = Label()
        label_widget.Text = label +":"
        label_widget.Dock = DockStyle.Right
        label_widget.AutoSize = True
        label_widget.Padding = Padding(2)

        input_widget = TextBox()
        input_widget.Text = str(default_value)
        input_widget.Dock = DockStyle.Right

        self.Controls.Add(label_widget)
        self.Controls.Add(input_widget)
        self.label = label_widget
        self.input = input_widget

    def resize_to_contents(self):
        self.Width = self.label.Width + self.input.Width

    def get_label(self):
        return self.label.Text[:-1]

    def get_value(self):
        return self.input.Text

    def get_value_as_int(self):
        return int(self.input.Text)

    def get_value_as_float(self):
        return float(self.input.Text)

# definition of folder picker
class FolderPicker(FlowLayoutPanel):

    def __init__(self, label, default_value=""):
        super(FolderPicker, self).__init__()
        self.Height = 30
        self.WrapContents = False
        self.default_value = default_value

        label_widget = Label()
        label_widget.Text = label +":"
        label_widget.AutoSize = True
        label_widget.Padding = Padding(5)

        input_widget = TextBox()
        input_widget.Text = str(default_value)
        input_widget.Dock = DockStyle.Right

        browse = Button()
        browse.Text = "Browse"
        browse.Height = 22
        browse.Click += self.get_picked_folder
        
        self.Controls.Add(label_widget)
        self.Controls.Add(input_widget)
        self.Controls.Add(browse)
        self.label = label_widget
        self.input = input_widget
        self.browse = browse

    def resize_to_contents(self):
        self.Width = self.label.Width + self.input.Width + self.browse.Width + 30
        pass

    def get_label(self):
        return self.label.Text[:-1]

    def get_value(self):
        return self.input.Text

    def get_picked_folder(self, *args):
        folder_picker = FolderBrowserDialog()
        folder_picker.Description = "Target Folder"
        if self.input.Text != "":
            folder_picker.SelectedPath = self.input.Text

        if folder_picker.ShowDialog():
            self.input.Text = folder_picker.SelectedPath        

class LabeledCheckBox(Panel):

    def __init__(self, label, default_value=True):
        super(LabeledCheckBox, self).__init__()
        self.Height = 30
        self.default_value = default_value

        label_widget = Label()
        label_widget.Text = label +":"
        label_widget.Dock = DockStyle.Right
        label_widget.AutoSize = True
        label_widget.Padding = Padding(7)

        input_widget = CheckBox()
        input_widget.Checked = default_value
        input_widget.Dock = DockStyle.Right
        input_widget.AutoSize = True

        self.Controls.Add(label_widget)
        self.Controls.Add(input_widget)
        self.label = label_widget
        self.input = input_widget

    def resize_to_contents(self):
        self.Width = self.label.Width

    def get_label(self):
        return self.label.Text[:-1]

    def get_value(self):
        return self.input.Checked

class StatePanel(Panel):

    def __init__(self, label, width, height):
        super(StatePanel, self).__init__()
        self.Width = width
        self.Height = height
        label_widget = Label()
        label_widget.Text = label
        label_widget.Dock = DockStyle.Top
        label_widget.Height = 20
        label_widget.BackColor = Color.Gray
        label_widget.Font = Font(label_widget.Font, FontStyle.Bold)
        label_widget.Padding = Padding(3)

        state_widget = RichTextBox()
        state_widget.Dock = DockStyle.Top
        state_widget.Height = height - label_widget.Height
        state_widget.BackColor = Color.Gray
        state_widget.Multiline = True
        state_widget.ScrollBars = RichTextBoxScrollBars.Vertical
        state_widget.WordWrap = True
        state_widget.ReadOnly = True


        self.Controls.Add(state_widget)
        self.Controls.Add(label_widget)
        self.label = label_widget
        self.info = state_widget

    def log(self, text, error=False):
        start_selection = self.info.Text.Length
        self.info.ReadOnly = False
        self.info.AppendText(str(text)+"\n")
        self.info.ReadOnly = True
        if error:
            # Change text color for error
            self.info.SelectionStart = start_selection
            self.info.SelectionLength = self.info.Text.Length - start_selection
            self.info.SelectionColor = Color.Red
        # Scroll down the Text box to the bottom
        self.info.SelectionStart = self.info.Text.Length
        self.info.ScrollToCaret()

    def log_status(self, text):
        self.label.Text = (str(text))

    def log_error(self, text):
        self.log(text, True)

    def clear_log(self):
        self.info.ReadOnly = False
        self.info.Text = ""
        self.info.ReadOnly = True

        
# definition of UI form        
class Alignment(Form):
    def __init__(self):
        self.Text = "No Polaris, no problem - polar alignment"
        self.TopMost = True
        self.Width = 722
        self.Height = 400
        self.BackColor = Color.FromName("Black")
        self.ForeColor = Color.FromName("Orange")

        #self.brush = System.Drawing.Brush(System.Drawing.Color.Black)
        
        # first row
        self.lat = ValueInput("Latitude", default_value=default_LAT)
        self.lat.Location = Point(-40, 10)
        self.lat.input.MaxLength = 6
        self.lat.input.Width = 40
        
        self.lon = ValueInput("Longitude", default_value=default_LON)
        self.lon.Location = Point(110, 10)
        self.lon.input.MaxLength = 6
        self.lon.input.Width = 40
        
        #self.testme = LabeledCheckBox("test mode", default_value=False)
        #self.testme.Location = Point(260, 10)
        
        self.testme = ComboInput("test mode", default_value="OFF")
        self.testme.Location = Point(260, 10)
        self.testme.input.Width = 70
        
        # second row
        self.pressure = ValueInput("Pressure", default_value=1013)
        self.pressure.Location = Point(-40, 40)
        self.pressure.input.MaxLength = 5
        self.pressure.input.Width = 40
        
        self.amb_temp = ValueInput("Ambient Temp", default_value=5)
        self.amb_temp.Location = Point(110, 40)
        self.amb_temp.input.MaxLength = 5
        self.amb_temp.input.Width = 40        

        
        # third row
        self.startALT = ValueInput("Starting ALT", default_value=15)
        self.startALT.Location = Point(-40, 70)
        self.startALT.input.MaxLength = 5
        self.startALT.input.Width = 40
        
        self.startAZ = ValueInput("Starting AZ", default_value=80)
        self.startAZ.Location = Point(110, 70)
        self.startAZ.input.MaxLength = 5
        self.startAZ.input.Width = 40
                
        self.endAZ = ValueInput("AZ end", default_value=175)
        self.endAZ.Location = Point(260, 70)
        self.endAZ.input.MaxLength = 5
        self.endAZ.input.Width = 40
        
        # fourth row
        self.no_of_pics = ValueInput("Number of pictures", default_value=10)
        self.no_of_pics.Location = Point(-40, 100)
        self.no_of_pics.input.MaxLength = 5
        self.no_of_pics.input.Width = 40
        
        # fifth row
        self.exposure = ValueInput("Exposure Time", default_value = default_exp)
        self.exposure.Location = Point(-40, 130)
        self.exposure.input.MaxLength = 5
        self.exposure.input.Width = 40
        
        self.gain = ValueInput("Gain", default_value = default_gain)
        self.gain.Location = Point(110, 130)
        self.gain.input.MaxLength = 5
        self.gain.input.Width = 40
        
        self.fov = ValueInput("FOV in Deg", default_value = default_FOV)
        self.fov.Location = Point(260, 130)
        self.fov.input.MaxLength = 5
        self.fov.input.Width = 40
 
        
        #self.FillEllipse(Brushes.Green, -40, 20, 150, 150)
        
        # Folder picker
        self.target_folder = FolderPicker("Output Folder", default_value = default_path)
        self.target_folder.Location = Point(10, 160)
        self.target_folder.input.Width = 276
        self.target_folder.resize_to_contents()

        # Process
        self.process = StatePanel("Process", width=210, height=232)
        self.process.Location = Point(480, 10)
        self.process.BorderStyle = BorderStyle.FixedSingle

        ###################################################################################
        ### First row of buttons
        
        # Button Position Update
        self.pos = Button()
        self.pos.Text = 'update ASCOM Position'
        self.pos.Width = 100
        self.pos.Height = 60
        self.pos.Location = Point(20, 190)
        self.pos.Font = Font(self.pos.Font, FontStyle.Bold)
        self.pos.Click += self.pos_start

        # Button Slew
        self.slew = Button()
        self.slew.Text = 'slew mount and capture'
        self.slew.Width = 100
        self.slew.Height = 60
        self.slew.Location = Point(134, 190)
        self.slew.Font = Font(self.slew.Font, FontStyle.Bold)
        self.slew.Click += self.slew_start
        
        # Button Solve
        self.solve = Button()
        self.solve.Text = 'platesolve results'
        self.solve.Width = 100
        self.solve.Height = 60
        self.solve.Location = Point(248, 190)
        self.solve.Font = Font(self.solve.Font, FontStyle.Bold)
        self.solve.Click += self.solve_start
        
        # Button analysis
        self.analysis = Button()
        self.analysis.Text = 'calculate correction angles'
        self.analysis.Width = 100
        self.analysis.Height = 60
        self.analysis.Location = Point(362, 190)
        self.analysis.Font = Font(self.solve.Font, FontStyle.Bold)
        self.analysis.Click += self.analysis_start
        
        ###################################################################################
        ### Second row of buttons
        
        # Button SE align
        self.SEalign = Button()
        self.SEalign.Text = 'Bright Star SE star sync'
        self.SEalign.Width = 100
        self.SEalign.Height = 60
        self.SEalign.Location = Point(20, 260)
        self.SEalign.Font = Font(self.solve.Font, FontStyle.Bold)
        self.SEalign.Click += self.SEalign_start
        
        # Button S align
        self.Salign = Button()
        self.Salign.Text = 'Bright Star S star sync'
        self.Salign.Width = 100
        self.Salign.Height = 60
        self.Salign.Location = Point(134, 260)
        self.Salign.Font = Font(self.solve.Font, FontStyle.Bold)
        self.Salign.Click += self.Salign_start
        
        # Button correct ALT
        self.corrALT = Button()
        self.corrALT.Text = 'Move for ALT corr'
        self.corrALT.Width = 100
        self.corrALT.Height = 60
        self.corrALT.Location = Point(248, 260)
        self.corrALT.Font = Font(self.solve.Font, FontStyle.Bold)
        self.corrALT.Click += self.corrALT_start
        
        # Button correct AZ
        self.corrAZ = Button()
        self.corrAZ.Text = 'Move for ALT and AZ corr'
        self.corrAZ.Width = 100
        self.corrAZ.Height = 60
        self.corrAZ.Location = Point(362, 260)
        self.corrAZ.Font = Font(self.solve.Font, FontStyle.Bold)
        self.corrAZ.Click += self.corrAZ_start
        
        ###################################################################################
        ### Labels, bottom right
        
        self.name_LAZ = Label()
        self.name_LAZ.Text= "Req. AZ Correction in DEG"
        self.name_LAZ.Location = Point(480,260)
        self.Controls.Add(self.name_LAZ)
        self.name_LAZ.Width = 150
    
        self.name_AZ = Label()
        self.name_AZ.Text=str(round(AZ_CORR_DEG, 2))
        self.name_AZ.Location = Point(640,260)
        self.Controls.Add(self.name_AZ)
        self.name_AZ.Width = 60


        self.name_LALT = Label()
        self.name_LALT.Text= "Req. ALT Correction in DEG"
        self.name_LALT.Location = Point(480,290)
        self.Controls.Add(self.name_LALT)
        self.name_LALT.Width = 150
        
        self.name_ALT = Label()
        self.name_ALT.Text=str(round(ALT_CORR_DEG, 2))
        self.name_ALT.Location = Point(640,290)
        self.Controls.Add(self.name_ALT)
        self.name_ALT.Width = 60

        ###################################################################################
        ### Controls

        self.Controls.Add(self.lat)
        self.Controls.Add(self.lon)
        self.Controls.Add(self.testme)
        
        self.Controls.Add(self.pressure)
        self.Controls.Add(self.amb_temp)
        
        self.Controls.Add(self.startALT)
        self.Controls.Add(self.startAZ)
        self.Controls.Add(self.endAZ)
        self.Controls.Add(self.no_of_pics)
        
        self.Controls.Add(self.exposure)
        self.Controls.Add(self.gain)
        self.Controls.Add(self.fov)
        
        self.Controls.Add(self.target_folder)
        self.Controls.Add(self.process)
        
        self.Controls.Add(self.pos)
        self.Controls.Add(self.slew)
        self.Controls.Add(self.solve)
        self.Controls.Add(self.analysis)
        
        self.Controls.Add(self.SEalign)
        self.Controls.Add(self.Salign)
        self.Controls.Add(self.corrALT)
        self.Controls.Add(self.corrAZ)
        
        self.processing = False
        self.ResumeLayout(False)

    def MainFormPaint(self, sender, e):
        graphics = e.Graphics       
        i = 0
        while i < count:
            graphics.FillEllipse(Brushes.White, 30*(i+1), 330, 20, 20)
            if color_ball[i] == 0:
                graphics.FillEllipse(Brushes.Red, 30*(i+1)+2, 332, 16, 16)
            if color_ball[i] == 1:
                graphics.FillEllipse(Brushes.Lime, 30*(i+1)+2, 332, 16, 16)
            i += 1
            
    
    def OnFormClosing(self, e, *args):
        e.Cancel = True
        form.Hide()
        
    def pos_start(self, *args):

        self.process.clear_log()

        # Get and convert values from the Windows form into usable types
        lat = self.get_float_from(self.lat)
        if lat is None: return
        
        lon = self.get_float_from(self.lon)
        if lon is None: return

        output = self.target_folder.get_value()
        #pause = self.pause.get_value()

        # Call the ASTAP solver command
        pos_loop(
            lat,
            lon,
            output,
            self
        )

    def slew_start(self, *args):
        if self.processing:
            global aborted
            aborted = True
            self.slew_end(aborted)
            return

        self.process.clear_log()

        # Get and convert values from the Windows form into usable types

        pressure = self.get_float_from(self.pressure)
        if pressure is None: return
        
        amb_temp = self.get_float_from(self.amb_temp)
        if amb_temp is None: return
        
        lat = self.get_float_from(self.lat)
        if lat is None: return

        startALT = self.get_int_from(self.startALT)
        if startALT is None: return

        startAZ = self.get_float_from(self.startAZ)
        if startAZ is None: return

        endAZ = self.get_float_from(self.endAZ)
        if endAZ is None: return
        
        no_of_pics = self.get_float_from(self.no_of_pics)
        if no_of_pics is None: return
        
        exposure = self.get_float_from(self.exposure)
        if exposure is None: return
        
        gain = self.get_float_from(self.gain)
        if gain is None: return
        
        testme = self.testme.get_value()
        
        output = self.target_folder.get_value()

        #pause = self.pause.get_value()

        # Call the capture command
        self.processing = True
        self.slew.Text = "Abort Alignment"
        slew_loop(
            lat,
            startALT,
            startAZ,
            endAZ,
            no_of_pics,
            exposure,
            gain,
            testme,
            output,            
            self
        )
        
    def solve_start(self, *args):
        if self.processing:
            global aborted
            aborted = True
            self.solve_end(aborted)
            return

        self.process.clear_log()

        # Get and convert values from the Windows form into usable types
        lat = self.get_float_from(self.lat)
        if lat is None: return        
        
        lon = self.get_float_from(self.lon)
        if lon is None: return
        
        pressure = self.get_float_from(self.pressure)
        if pressure is None: return
        
        amb_temp = self.get_float_from(self.amb_temp)
        if amb_temp is None: return
        
        fov = self.get_float_from(self.fov)
        if fov is None: return
        
        testme = self.testme.get_value()
        
        output = self.target_folder.get_value()        

        #pause = self.pause.get_value()

        # Call the ASTAP solver command
        self.processing = True
        self.solve.Text = "platesolve results"
        solve_loop(
            lat,
            lon,
            pressure,
            amb_temp,
            fov,
            testme,
            output,            
            self
        )
        
    
    def analysis_start(self, *args):

        self.process.clear_log()

        # Get and convert values from the Windows form into usable types
        lat = self.get_float_from(self.lat)
        if lat is None: return
        
        lon = self.get_float_from(self.lon)
        if lon is None: return

        output = self.target_folder.get_value()
        #pause = self.pause.get_value()

        # Call the ASTAP solver command
        self.processing = True
        analysis_loop(
            lat,
            lon,
            output,
            self
        )    
    
    #######################################################################
    ##### Start of lower button row
    def SEalign_start(self, *args):
        if self.processing:
            global aborted
            aborted = True
            self.SEalign_end(aborted)
            return

        self.process.clear_log()

        # Get and convert values from the Windows form into usable types
        lat = self.get_float_from(self.lat)
        if lat is None: return
        
        lon = self.get_float_from(self.lon)
        if lon is None: return

        output = self.target_folder.get_value()
        #pause = self.pause.get_value()

        # Go into the SEalign_loop
        self.processing = True
        self.SEalign.Text = "Abort Sync"
        SEalign_loop(
            lat,
            lon,
            output,
            self
        )    
    
    
    def Salign_start(self, *args):
        if self.processing:
            global aborted
            aborted = True
            self.SEalign_end(aborted)
            return

        self.process.clear_log()

        # Get and convert values from the Windows form into usable types
        lat = self.get_float_from(self.lat)
        if lat is None: return
        
        lon = self.get_float_from(self.lon)
        if lon is None: return

        output = self.target_folder.get_value()
        #pause = self.pause.get_value()

        # Go into the Salign_loop
        self.processing = True
        self.Salign.Text = "Abort Sync"
        Salign_loop(
            lat,
            lon,
            output,
            self
        )    
    
    
    def corrALT_start(self, *args):
        if self.processing:
            global aborted
            aborted = True
            self.corrALT_end(aborted)
            return

        self.process.clear_log()

        # Get and convert values from the Windows form into usable types
        lat = self.get_float_from(self.lat)
        if lat is None: return
        
        lon = self.get_float_from(self.lon)
        if lon is None: return

        output = self.target_folder.get_value()
        #pause = self.pause.get_value()

        # Call the ASTAP solver command
        self.processing = True
        self.corrALT.Text = "Abort Correction"
        corrALT_loop(
            lat,
            lon,
            output,
            self
        )    
    
    
    def corrAZ_start(self, *args):
        if self.processing:
            global aborted
            aborted = True
            self.corrAZ_end(aborted)
            return

        self.process.clear_log()

        # Get and convert values from the Windows form into usable types
        lat = self.get_float_from(self.lat)
        if lat is None: return
        
        lon = self.get_float_from(self.lon)
        if lon is None: return

        output = self.target_folder.get_value()
        #pause = self.pause.get_value()

        # Call the ASTAP solver command
        self.processing = True
        self.corrAZ.Text = "Abort Correction"
        corrAZ_loop(
            lat,
            lon,
            output,
            self
        )        
  
    
    def pos_paint(self, sender, e):
        graphics = e.Graphics
        global readyness
        
        self.name_readyness = Label()
        self.name_readyness.Text= "Readyness:"
        self.name_readyness.Location = Point(480,320)
        self.Controls.Add(self.name_readyness)
        self.name_readyness.Width = 100
        
        if readyness == True:
            graphics.FillEllipse(Brushes.Green, 650, 320, 20, 20)
        else:
            graphics.FillEllipse(Brushes.Red, 650, 320, 20, 20)    
    
    
    def pos_end(self):
        self.Paint += self.pos_paint
        self.Refresh()
    

    def slew_end(self, aborted=False):
        self.processing = False
        self.slew.Text = "Slew mount and capture"
        if aborted:
            SharpCap.Mounts.SelectedMount.Stop()
            self.process.log_error("Alignment Aborted!")
            self.process.log_error("Mount Stopped!")
            
            
    def slew_finish(self):
        self.Paint += self.slew_paint
        self.Refresh()
        
        
    def slew_paint(self, sender, e):
        graphics = e.Graphics       
        i = 0
        while i < count:
            graphics.FillEllipse(Brushes.White, 30*(i+1), 330, 20, 20)
            i += 1


    def solve_end(self, aborted=False):
        self.processing = False
        self.solve.Text = "Platesolve results"
        if aborted:
            self.process.log_error("Platesolving Aborted!")


    def SEalign_end(self, aborted=False):
        self.processing = False
        self.SEalign.Text = "Bright Star SE star sync"
        if aborted:
            SharpCap.Mounts.SelectedMount.Stop()
            self.process.log_error("Bright Star sync Aborted!")
            self.process.log_error("Mount Stopped!")


    def Salign_end(self, aborted=False):
        self.processing = False
        self.Salign.Text = "Bright Star S star sync"
        if aborted:
            SharpCap.Mounts.SelectedMount.Stop()
            self.process.log_error("Bright Star sync Aborted!")
            self.process.log_error("Mount Stopped!")


    def corrALT_end(self, aborted=False):
        self.processing = False
        self.corrALT.Text = "Move for ALT corr"
        if aborted:
            SharpCap.Mounts.SelectedMount.Stop()
            self.process.log_error("ALT correction move Aborted!")
            self.process.log_error("Mount Stopped!")


    def corrAZ_end(self, aborted=False):
        self.processing = False
        self.corrAZ.Text = "Move for ALT and AZ corr"
        if aborted:
            SharpCap.Mounts.SelectedMount.Stop()
            self.process.log_error("ALT and AZ correction move Aborted!")
            self.process.log_error("Mount Stopped!")


    def get_int_from(self, attribute):
        try:
            return attribute.get_value_as_int()

        except:
            self.process.log_error(
                "[Error] Wrong value type for '%s', expected integer: %s"%\
                            (attribute.get_label(), attribute.get_value())
            )
            return None


    def get_float_from(self, attribute):
        try:
            return attribute.get_value_as_float()

        except:
            self.process.log_error(
                "[Error] Wrong value type for '%s', expected float: %s"%\
                            (attribute.get_label(), attribute.get_value())
            )
            return None


    def paint_alert(self):
        self.BackColor = Color.FromName("Black")
    
    
    def paint_finish(self):
        self.Paint += self.MainFormPaint
        self.Refresh()


    def analysis_finish(self):
        self.processing = False
        self.name_AZ.Text=str(round(AZ_CORR_DEG, 2))
        self.name_ALT.Text=str(round(ALT_CORR_DEG, 2)) 

            
form = Alignment()
form.icon = Icon(default_icon)


def show_script():
    form.Show()

SharpCap.AddCustomButton("POLAR ALIGN", Image.FromFile(default_icon), None, show_script)        
        

        
        
        
        
        